/* -*- buffer-read-only: t -*- vi: set ro:
#line 636
   THIS FILE IS GENERATED AUTOMATICALLY.  PLEASE DO NOT EDIT.
#line 636
*/
#define MF_SOURCE_NAME  MF_SOURCE_BI_DB
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include "mailfromd.h"
#include "prog.h"
#include "builtin.h"

#line 15 "bi_db.m4"
/* This file is part of Mailfromd.             -*- c -*-
#line 15
   Copyright (C) 2006, 2007, 2008, 2009 Sergey Poznyakoff
#line 15

#line 15
   This program is free software; you can redistribute it and/or modify
#line 15
   it under the terms of the GNU General Public License as published by
#line 15
   the Free Software Foundation; either version 3, or (at your option)
#line 15
   any later version.
#line 15

#line 15
   This program is distributed in the hope that it will be useful,
#line 15
   but WITHOUT ANY WARRANTY; without even the implied warranty of
#line 15
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#line 15
   GNU General Public License for more details.
#line 15

#line 15
   You should have received a copy of the GNU General Public License
#line 15
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define DEFAULT_DB_MODE 0640
#include <fnmatch.h>

struct db_prop {                /* Database properties */
	char *pat;              /* Database name pattern */
	mode_t mode;            /* File mode */
	int null;               /* Null byte */
};

static mu_list_t db_prop_list;

static int
db_prop_compare(const void *item, const void *data)
{
	const struct db_prop *prop = item;
	const char *name = data;
	return strcmp(prop->pat, name);
}

static int
strtomode(char *str, mode_t *pmode)
{
	mode_t mode = 0;
	struct { char c; unsigned mask; } modetab[] = {
		{ 'r', S_IRUSR },
		{ 'w', S_IWUSR },
		{ 'x', S_IXUSR },
		{ 'r', S_IRGRP },
		{ 'w', S_IWGRP },
		{ 'x', S_IXGRP },
		{ 'r', S_IROTH },
		{ 'w', S_IWOTH },
		{ 'x', S_IXOTH },
		{ 0 }
	};
	int i;

	for (i = 0; modetab[i].c; i++) {
		if (!str[i])
			return i + 1;
		if (str[i] == modetab[i].c)
			mode |= modetab[i].mask;
		else if (str[i] != '-')
			return i + 1;
	}
	if (str[i])
		return i + 1;
	*pmode = mode;
	return 0;
}

/* #pragma dbprop <name> [null] [mode] */
#line 69 "bi_db.m4"

#line 69
static void _pragma_dbprop (int argc, char **argv, const char *text)
#line 69

{
	int null = 0;
	mode_t mode = DEFAULT_DB_MODE;
	struct db_prop *prop;
	int rc;
	char *pat;

	--argc;
	pat = *++argv;
	
	while (--argc) {
		char *p = *++argv;

		if (strcmp (p, "null") == 0)
			null = 1;
		else if (c_isdigit(*p)) {
			unsigned long n = strtoul(p, &p, 8);
			if (*p || (mode = n) != n) {
				parse_error(_("bad numeric file mode"));
				return;
			}
		} else if (rc = strtomode(p, &mode)) {
			parse_error(_("bad symbolic file mode (near %s)"),
				    p + rc - 1);
			return;
		}
	}

	if (!db_prop_list) {
		rc = mu_list_create(&db_prop_list);
		if (rc) {
			parse_error(_("cannot create list: %s"),
				    mu_strerror(rc));
			return;
		}
		mu_list_set_comparator(db_prop_list, db_prop_compare);
	}

	if (mu_list_locate(db_prop_list, pat, (void**) &prop)) {
		prop = xmalloc(sizeof(*prop));
		prop->pat = xstrdup(pat);
		rc = mu_list_append(db_prop_list, prop);
		if (rc) {
			parse_error(_("Cannot create list: %s"),
				    mu_strerror(rc));
			return;
		}
	}
	prop->mode = mode;
	prop->null = null;
}

const struct db_prop *
db_prop_lookup(const char *name)
{
	mu_iterator_t itr;
	const struct db_prop *found = NULL;
	
	if (db_prop_list
	    && mu_list_get_iterator(db_prop_list, &itr) == 0) {
		for (mu_iterator_first(itr);
		     !mu_iterator_is_done(itr);
		     mu_iterator_next(itr)) {
			const struct db_prop *p;
			mu_iterator_current(itr, (void**)&p);
			if (strcmp(p->pat, name) == 0) {
				found = p;
				break;
			} else if (fnmatch(p->pat, name, 0) == 0)
				found = p;
		}
		mu_iterator_destroy(&itr);
	}
	return found;	
}


#define LOOKUP_NULL_BYTE 0x1
#define LOOKUP_TEST_ONLY 0x2

static int
dbmap_lookup(eval_environ_t env, char *dbname, const char *keystr,
	     const char *defval, int flags)
{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;

	if (!defval)
		defval = "";
	if (mu_dbm_open(dbname, &db, MU_STREAM_READ, 0, NULL))
			env_throw_bi(env, mfe_dbfailure, NULL, _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 166

	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = (void*) keystr;
	MU_DATUM_SIZE(key) = strlen(keystr);
	if (flags & LOOKUP_NULL_BYTE)
		MU_DATUM_SIZE(key)++;
	rc = mu_dbm_fetch(&db, key, &contents) == 0;
	debug2(10, "Looking up %s: %s", keystr, rc ? "true" : "false");
	if (flags & LOOKUP_TEST_ONLY) 
		push(env, (STKVAL)rc);
	else {
		if (!rc) {
			if (defval)
				pushs(env, (STKVAL)defval);
			else
				push(env, 0);
		} else if (((char*)MU_DATUM_PTR(contents))[MU_DATUM_SIZE(contents)-1]) {
			size_t off;
			size_t len = MU_DATUM_SIZE(contents);
			char *s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
			memcpy(s, MU_DATUM_PTR(contents), len);
			s[len] = 0;
			push(env, (STKVAL) off);
		} else
			pushs(env, MU_DATUM_PTR(contents));
	}
	mu_dbm_datum_free(&contents);
	mu_dbm_close(&db);
	return rc;
}

void
#line 198
bi_dbmap(eval_environ_t env)
#line 198

#line 198 "bi_db.m4"
{
#line 198
	
#line 198

#line 198
long __bi_argcnt;
#line 198
char * dbname;
#line 198
        char * key;
#line 198
        long  null;
#line 198
        
#line 198
get_string_arg(env, 1, &dbname);
#line 198
        get_string_arg(env, 2, &key);
#line 198
        get_numeric_arg(env, 3, &null);
#line 198
        
#line 198
get_numeric_arg(env, 0, &__bi_argcnt);
#line 198
        adjust_stack(env, __bi_argcnt + 1);
#line 198

#line 198
	if (prog_trace_option & BUILTIN_MASK_db)
#line 198
		prog_trace(env, "dbmap %s %s %lu",dbname, key, ((__bi_argcnt > 2) ? null : 0));;
#line 198

{
	const struct db_prop *prop = db_prop_lookup(dbname);
	dbmap_lookup(env, dbname, key, NULL,
		     LOOKUP_TEST_ONLY | 
		         (((__bi_argcnt > 2) ? null : prop && prop->null)
			   ? LOOKUP_NULL_BYTE : 0));
}
}

void
#line 208
bi_dbget(eval_environ_t env)
#line 208

#line 208 "bi_db.m4"
{
#line 208
	
#line 208

#line 208
long __bi_argcnt;
#line 208
char * dbname;
#line 208
        char * key;
#line 208
        char * defval;
#line 208
        long  null;
#line 208
        
#line 208
get_string_arg(env, 1, &dbname);
#line 208
        get_string_arg(env, 2, &key);
#line 208
        get_string_arg(env, 3, &defval);
#line 208
        get_numeric_arg(env, 4, &null);
#line 208
        
#line 208
get_numeric_arg(env, 0, &__bi_argcnt);
#line 208
        adjust_stack(env, __bi_argcnt + 1);
#line 208

#line 208
	if (prog_trace_option & BUILTIN_MASK_db)
#line 208
		prog_trace(env, "dbget %s %s %s %lu",dbname, key, ((__bi_argcnt > 2) ? defval : ""), ((__bi_argcnt > 3) ? null : 0));;

{
	const struct db_prop *prop = db_prop_lookup(dbname);
	dbmap_lookup(env, dbname, key, 
                     ((__bi_argcnt > 2) ? defval : 0), 
		     ((__bi_argcnt > 3) ? null : prop && prop->null)
		      ? LOOKUP_NULL_BYTE : 0);
}
}

void
#line 219
bi_dbput(eval_environ_t env)
#line 219

#line 219 "bi_db.m4"
{
#line 219
	
#line 219

#line 219
long __bi_argcnt;
#line 219
char * dbname;
#line 219
        char * keystr;
#line 219
        char * value;
#line 219
        long  null;
#line 219
        long  mode;
#line 219
        
#line 219
get_string_arg(env, 1, &dbname);
#line 219
        get_string_arg(env, 2, &keystr);
#line 219
        get_string_arg(env, 3, &value);
#line 219
        get_numeric_arg(env, 4, &null);
#line 219
        get_numeric_arg(env, 5, &mode);
#line 219
        
#line 219
get_numeric_arg(env, 0, &__bi_argcnt);
#line 219
        adjust_stack(env, __bi_argcnt + 1);
#line 219

#line 219
	if (prog_trace_option & BUILTIN_MASK_db)
#line 219
		prog_trace(env, "dbput %s %s %s %lu %lu",dbname, keystr, value, ((__bi_argcnt > 3) ? null : 0), ((__bi_argcnt > 4) ? mode : 0));;

{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_RDWR,
			((__bi_argcnt > 4) ? mode : (prop ? prop->mode : DEFAULT_DB_MODE)),
			NULL))
			env_throw_bi(env, mfe_dbfailure, "dbput", _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 235
	memset(&key, 0, sizeof key);
	MU_DATUM_PTR(key) = keystr;
	MU_DATUM_SIZE(key) = strlen(keystr);
	if (((__bi_argcnt > 3) ? null : prop && prop->null))
		MU_DATUM_SIZE(key)++;

	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(contents) = value;
	MU_DATUM_SIZE(contents) = strlen(value) + 1;
	
	rc = mu_dbm_insert(&db, key, contents, 1);
	mu_dbm_close(&db);
		if (!(rc == 0))
#line 247
			env_throw_bi(env, mfe_dbfailure, "dbput", _("failed to insert data to %s: %s %s: %s"),dbname,keystr,value,mu_dbm_strerror())
#line 253
;
}
}

void
#line 257
bi_dbdel(eval_environ_t env)
#line 257

#line 257 "bi_db.m4"
{
#line 257
	
#line 257

#line 257
long __bi_argcnt;
#line 257
char * dbname;
#line 257
        char * keystr;
#line 257
        long  null;
#line 257
        long  mode;
#line 257
        
#line 257
get_string_arg(env, 1, &dbname);
#line 257
        get_string_arg(env, 2, &keystr);
#line 257
        get_numeric_arg(env, 3, &null);
#line 257
        get_numeric_arg(env, 4, &mode);
#line 257
        
#line 257
get_numeric_arg(env, 0, &__bi_argcnt);
#line 257
        adjust_stack(env, __bi_argcnt + 1);
#line 257

#line 257
	if (prog_trace_option & BUILTIN_MASK_db)
#line 257
		prog_trace(env, "dbdel %s %s %lu %lu",dbname, keystr, ((__bi_argcnt > 2) ? null : 0), ((__bi_argcnt > 3) ? mode : 0));;

{
	DBM_FILE db;
	DBM_DATUM key;
	int rc;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_RDWR,
			((__bi_argcnt > 3) ? mode : (prop ? prop->mode : DEFAULT_DB_MODE)),
			NULL))
			env_throw_bi(env, mfe_dbfailure, "dbdel", _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 272
	memset(&key, 0, sizeof key);
	MU_DATUM_PTR(key) = keystr;
	MU_DATUM_SIZE(key) = strlen(keystr);
	if (((__bi_argcnt > 2) ? null : prop && prop->null))
		MU_DATUM_SIZE(key)++;
	rc = mu_dbm_delete(&db, key);
	mu_dbm_close(&db);
		if (!(rc == 0))
#line 279
			env_throw_bi(env, mfe_dbfailure, "dbdel", _("failed to delete data `%s' from `%s': %s"),keystr,dbname,mu_dbm_strerror())
#line 284
;
}
}


#define NUMDB 128

struct db_tab {
	int used;
	DBM_FILE db;
	DBM_DATUM key;
};

static void *
alloc_db_tab()
{
	return xcalloc(NUMDB, sizeof(struct db_tab));
}

static void
close_db_tab(struct db_tab *dbt)
{
	if (dbt->used) {
		mu_dbm_datum_free(&dbt->key);
		mu_dbm_close(&dbt->db);
		dbt->used = 0;
	}
}

static void
destroy_db_tab(void *data)
{
	int i;
	struct db_tab *db = data;
	for (i = 0; i < NUMDB; i++) 
		close_db_tab(db + i);
	free(db);
}


#line 323

#line 323
static int DBTAB_id;
#line 323 "bi_db.m4"
;

static int
new_db_tab(struct db_tab *dbt)
{
	int i;
	for (i = 0; i < NUMDB; i++)
		if (!dbt[i].used) {
			dbt[i].used = 1;
			return i;
		}
	return -1;
}
			


void
#line 339
bi_dbfirst(eval_environ_t env)
#line 339

#line 339 "bi_db.m4"
{
#line 339
	
#line 339

#line 339

#line 339
char * dbname;
#line 339
        
#line 339
get_string_arg(env, 0, &dbname);
#line 339
        
#line 339

#line 339
        adjust_stack(env, 1);
#line 339

#line 339
	if (prog_trace_option & BUILTIN_MASK_db)
#line 339
		prog_trace(env, "dbfirst %s",dbname);;
#line 339

{
	int rc;
	int n;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id);
	DBM_FILE db;
	DBM_DATUM key;
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_READ, 0, NULL))
			env_throw_bi(env, mfe_dbfailure, "dbfirst", _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 352
	rc = mu_dbm_firstkey(&db, &key);
		if (!(rc == 0))
#line 353
			env_throw_bi(env, mfe_dbfailure, "dbfirst", _("mu_dbm_firstkey failed: %s"),mu_dbm_strerror())
#line 355
;
	n = new_db_tab(dbt);
		if (!(n >= 0))
#line 357
			env_throw_bi(env, mfe_failure, "dbfirst", _("no more database entries available"))
#line 359
;
	dbt += n;
	dbt->db = db;
	dbt->key = key;
	
#line 363
do {
#line 363
  push(env, (STKVAL) (n));
#line 363
  return;
#line 363
} while (0);
}
}

void
#line 367
bi_dbnext(eval_environ_t env)
#line 367

#line 367 "bi_db.m4"
{
#line 367
	
#line 367

#line 367

#line 367
long  dn;
#line 367
        
#line 367
get_numeric_arg(env, 0, &dn);
#line 367
        
#line 367

#line 367
        adjust_stack(env, 1);
#line 367

#line 367
	if (prog_trace_option & BUILTIN_MASK_db)
#line 367
		prog_trace(env, "dbnext %lu",dn);;
#line 367

{
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	DBM_DATUM nextkey;
	int rc;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 373
			env_throw_bi(env, mfe_range, "dbnext", _("invalid database descriptor"))
#line 375
;

	rc = mu_dbm_nextkey(&dbt->db, dbt->key, &nextkey);
	if (rc) {
		close_db_tab(dbt);
		
#line 380
do {
#line 380
  push(env, (STKVAL) (0));
#line 380
  return;
#line 380
} while (0);
	}
	mu_dbm_datum_free(&dbt->key);
	dbt->key = nextkey;
	
#line 384
do {
#line 384
  push(env, (STKVAL) (1));
#line 384
  return;
#line 384
} while (0);
}
}

void
#line 388
bi_dbkey(eval_environ_t env)
#line 388

#line 388 "bi_db.m4"
{
#line 388
	
#line 388

#line 388

#line 388
long  dn;
#line 388
        
#line 388
get_numeric_arg(env, 0, &dn);
#line 388
        
#line 388

#line 388
        adjust_stack(env, 1);
#line 388

#line 388
	if (prog_trace_option & BUILTIN_MASK_db)
#line 388
		prog_trace(env, "dbkey %lu",dn);;
#line 388

{
	size_t off, len;
	char *s;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 394
			env_throw_bi(env, mfe_range, "dbkey", _("invalid database descriptor"))
#line 396
;
	
	len = MU_DATUM_SIZE(dbt->key);
	s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
	memcpy(s, MU_DATUM_PTR(dbt->key), len);
	s[len] = 0;
	
#line 402
do {
#line 402
  push(env, (STKVAL) (off));
#line 402
  return;
#line 402
} while (0);
}
}

void
#line 406
bi_dbvalue(eval_environ_t env)
#line 406

#line 406 "bi_db.m4"
{
#line 406
	
#line 406

#line 406

#line 406
long  dn;
#line 406
        
#line 406
get_numeric_arg(env, 0, &dn);
#line 406
        
#line 406

#line 406
        adjust_stack(env, 1);
#line 406

#line 406
	if (prog_trace_option & BUILTIN_MASK_db)
#line 406
		prog_trace(env, "dbvalue %lu",dn);;
#line 406

{
	int rc;
	size_t off, len;
	char *s;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	DBM_DATUM contents;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 414
			env_throw_bi(env, mfe_range, "dbvalue", _("invalid database descriptor"))
#line 416
;

	memset(&contents, 0, sizeof contents);
	rc = mu_dbm_fetch(&dbt->db, dbt->key, &contents);
		if (!(rc == 0))
#line 420
			env_throw_bi(env, mfe_dbfailure, "dbvalue", _("key not found: %s"),mu_dbm_strerror())
#line 423
;
	
	len = MU_DATUM_SIZE(contents);
	s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
	memcpy(s, MU_DATUM_PTR(contents), len);
	s[len] = 0;
	mu_dbm_datum_free(&contents);
	
#line 430
do {
#line 430
  push(env, (STKVAL) (off));
#line 430
  return;
#line 430
} while (0);
}
}


enum greylist_semantics
{
	greylist_traditional,
	greylist_ct
};

static enum greylist_semantics greylist_semantics = greylist_traditional;

/* #pragma greylist {traditional|gray|ct|con-tassios}*/
#line 444 "bi_db.m4"

#line 444
static void _pragma_greylist (int argc, char **argv, const char *text)
#line 444

{
	if (strcmp(argv[1], "traditional") == 0
	    || strcmp(argv[1], "gray") == 0)
		greylist_semantics = greylist_traditional;
	else if (strcmp(argv[1], "ct") == 0
		 || strcmp(argv[1], "con-tassios") == 0)
		greylist_semantics = greylist_ct;
	else
		/* TRANSLATORS: Do not translate keywords:
		   traditional, gray, ct, con-tassios */
		parse_error(_("unknown semantics; allowed values are: "
			      "traditional (or gray) and "
			      "ct (or con-tassios)"));
}

static void
greylist_print_item(const char *key, size_t size, const void *content)
{
	time_t timestamp = *(time_t*) content;
	size--; /* Size includes the trailing nul */
	printf("%*.*s ", size, size, key);
	format_time_str(stdout, timestamp);
	putchar('\n');
}

static int
greylist_expire_item(const void *content)
{
	time_t timestamp = *(time_t*) content;
	return greylist_format->expire_interval &&
		time(NULL) - timestamp > greylist_format->expire_interval;
}


static struct db_format greylist_format_struct = {
	"greylist",
	DEFAULT_GREYLIST_DATABASE,
	1,
	DEFAULT_EXPIRE_INTERVAL,
	greylist_print_item,
	greylist_expire_item
};

struct db_format *greylist_format = &greylist_format_struct;

static size_t greylist_seconds_left_loc
#line 490 "bi_db.m4"
;

#line 494
/* The traditional (aka gray's) greylist implementation: the greylist
#line 494
   database keeps the time the greylisting was activated.
#line 494
 */
static int
do_greylist_traditional(eval_environ_t env, char *email, long interval)
{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	int readonly;
	time_t now;

	rc = mu_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600,
			 &readonly);
        	if (!(rc == 0))
#line 507
			env_throw_bi(env, mfe_dbfailure, NULL, _("mu_dbm_open(%s) failed: %s"),greylist_format->dbname,mu_dbm_strerror())
;
	
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = email;
	MU_DATUM_SIZE(key) = strlen(email)+1;

	time(&now);
	if (mu_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp, diff;

			if (!(MU_DATUM_SIZE(contents) == sizeof timestamp))
#line 519
			env_throw_bi(env, mfe_dbfailure, NULL, _("greylist database %s has wrong data size"),greylist_format->dbname)
#line 522
;
		
		timestamp = *(time_t*) MU_DATUM_PTR(contents);
		diff = now - timestamp;

		__DBG(20) {
			char timebuf[32];
			debug_log("%s entered greylist database on %s, "
				  "%ld seconds ago",
				  email,
				  mailfromd_timestr(timestamp, timebuf,
						    sizeof timebuf),
				  (long) diff);
		}

		if (diff < interval) {
			diff = interval - diff;

			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (diff);
			
			debug2(20, "%s still greylisted (for %lu sec.)",
			       email,
			       (unsigned long) diff);
			rc = 1;
		} else if (diff > greylist_format->expire_interval) {
			debug1(20, "greylist record for %s expired",
			       email);
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
			if (!readonly) {
				memcpy(MU_DATUM_PTR(contents),
				       &now, sizeof now);
				if (mu_dbm_insert(&db, key, contents, 1))
					mu_error(_("cannot insert datum `%-.*s' in "
						   "greylist database %s: %s"),
			                         MU_DATUM_SIZE(key),
						 (char*)MU_DATUM_PTR(key),
						 greylist_format->dbname,
	                                         mu_dbm_strerror());
			} else
				debug(20, "database opened in readonly mode: "
				      "not updating");
			rc = 1;
		} else {
			debug1(20, "%s finished greylisting period",
			       email);
			rc = 0;
		}
		mu_dbm_datum_free(&contents);
	} else if (!readonly) {
		debug1(20, "greylisting %s", email);
		*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
		MU_DATUM_PTR(contents) = (void*)&now;
		MU_DATUM_SIZE(contents) = sizeof now;
		if (mu_dbm_insert(&db, key, contents, 1))
			mu_error(_("Cannot insert datum `%-.*s' in greylist "
				   "database %s: %s"),
		                 MU_DATUM_SIZE(key), (char*)MU_DATUM_PTR(key),
				 greylist_format->dbname,
	                         mu_dbm_strerror());
		rc = 1;
	} else
		rc = 0;

	mu_dbm_close(&db);

	return rc;
}

#line 594
/* Implementation of the is_greylisted predicate has no sense for 
#line 594
   traditional greylist databases, because greylisting interval is
#line 594
   not known beforehand.
#line 594
   FIXME: keep the reference below up to date.
#line 594
 */
static int
is_greylisted_traditional(eval_environ_t env, char *email)
{
		env_throw_bi(env, mfe_failure, NULL, _("is_greylisted is not implemented for traditional greylist databases; "
#line 598
		   "see documentation, chapter %s %s for more info"),"4.12.1.23","Greylisting functions");
#line 602
	return 0;
}

#line 608
/* New greylist implementation (by Con Tassios): the database keeps
#line 608
   the time the greylisting period is set to expire (`interval' seconds
#line 608
   from now)
#line 608
*/
static int
do_greylist_ct(eval_environ_t env, char *email, long interval)
{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	int readonly;
	time_t now;

	rc = mu_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600,
			 &readonly);
        	if (!(rc == 0))
#line 621
			env_throw_bi(env, mfe_dbfailure, NULL, _("mu_dbm_open(%s) failed: %s"),greylist_format->dbname,mu_dbm_strerror())
;
	
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = email;
	MU_DATUM_SIZE(key) = strlen(email)+1;

	time(&now);
	if (mu_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp;

			if (!(MU_DATUM_SIZE(contents) == sizeof timestamp))
#line 633
			env_throw_bi(env, mfe_dbfailure, NULL, _("greylist database %s has wrong data size"),greylist_format->dbname)
#line 636
;
		
		timestamp = *(time_t*) MU_DATUM_PTR(contents);

		if (now < timestamp) {
			time_t diff = timestamp - now;
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (diff);
			
			debug2(20, "%s still greylisted (for %lu sec.)",
			       email,
			       (unsigned long) diff);
			rc = 1;
		} else if (now - timestamp >
			   greylist_format->expire_interval) {
			debug1(20, "greylist record for %s expired", email);
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
			if (!readonly) {
				now += interval;
				memcpy(MU_DATUM_PTR(contents),
				       &now, sizeof now);
				if (mu_dbm_insert(&db, key, contents, 1))
					mu_error(_("Cannot insert datum "
						   "`%-.*s' in greylist "
						   "database %s: %s"),
			                         MU_DATUM_SIZE(key),
						 (char*)MU_DATUM_PTR(key),
						 greylist_format->dbname,
	                                         mu_dbm_strerror());
			} else
				debug(20, "database opened in readonly mode: "
				      "not updating");
			rc = 1;
		} else {
			debug1(20, "%s finished greylisting period", email);
			rc = 0;
		}
		mu_dbm_datum_free(&contents);
	} else if (!readonly) {
		debug1(20, "greylisting %s", email);
		*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
		now += interval;
		MU_DATUM_PTR(contents) = (void*)&now;
		MU_DATUM_SIZE(contents) = sizeof now;
		if (mu_dbm_insert(&db, key, contents, 1))
			mu_error(_("Cannot insert datum `%-.*s' in greylist "
				   "database %s: %s"),
		                 MU_DATUM_SIZE(key), (char*)MU_DATUM_PTR(key),
				 greylist_format->dbname,
	                         mu_dbm_strerror());
		rc = 1;
	} else
		rc = 0;
	
	mu_dbm_close(&db);
	
	return rc;
}

/* The `is_greylisted' predicate for new databases */
static int
is_greylisted_ct(eval_environ_t env, char *email)
{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	int readonly;
	time_t now;

	rc = mu_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600,
			 &readonly);
		if (!(rc == 0))
#line 707
			env_throw_bi(env, mfe_dbfailure, NULL, _("mu_dbm_open(%s) failed: %s"),greylist_format->dbname,mu_dbm_strerror())
;

	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = email;
	MU_DATUM_SIZE(key) = strlen(email) + 1;

	time(&now);
	if (mu_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp;

			if (!(MU_DATUM_SIZE(contents) == sizeof timestamp))
#line 719
			env_throw_bi(env, mfe_dbfailure, NULL, _("greylist database %s has wrong data size"),greylist_format->dbname)
#line 722
;
                
		timestamp = *(time_t*) MU_DATUM_PTR(contents);
                
		rc = timestamp > now;
                if (rc)
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (timestamp - now);

		mu_dbm_datum_free(&contents);
	} else  
		rc = 0;

	mu_dbm_close(&db);

	return rc;
}

struct greylist_class {
	int (*gl_fun)(eval_environ_t, char *, long);
	int (*gl_pred)(eval_environ_t, char *);
};

struct greylist_class greylist_class[] = {
	{ do_greylist_traditional, is_greylisted_traditional },
	{ do_greylist_ct, is_greylisted_ct }
};

#line 753
/* greylist(key, interval)
#line 753

#line 753
   Returns true if the key is greylisted, false if it's OK to
#line 753
   deliver mail.
#line 753
 */
void
#line 754
bi_greylist(eval_environ_t env)
#line 754

#line 754 "bi_db.m4"
{
#line 754
	
#line 754

#line 754

#line 754
char * email;
#line 754
        long  interval;
#line 754
        
#line 754
get_string_arg(env, 0, &email);
#line 754
        get_numeric_arg(env, 1, &interval);
#line 754
        
#line 754

#line 754
        adjust_stack(env, 2);
#line 754

#line 754
	if (prog_trace_option & BUILTIN_MASK_db)
#line 754
		prog_trace(env, "greylist %s %lu",email, interval);;
#line 754

{
	
#line 756
do {
#line 756
  push(env, (STKVAL) (greylist_class[greylist_semantics].gl_fun(env, email,
#line 756
							    interval)));
#line 756
  return;
#line 756
} while (0);
#line 758
}
}

#line 765
/* is_greylisted(key)
#line 765
   
#line 765
   Returns true if the key is greylisted, otherwise false
#line 765
   
#line 765
*/
void
#line 766
bi_is_greylisted(eval_environ_t env)
#line 766

#line 766 "bi_db.m4"
{
#line 766
	
#line 766

#line 766

#line 766
char * email;
#line 766
        
#line 766
get_string_arg(env, 0, &email);
#line 766
        
#line 766

#line 766
        adjust_stack(env, 1);
#line 766

#line 766
	if (prog_trace_option & BUILTIN_MASK_db)
#line 766
		prog_trace(env, "is_greylisted %s",email);;
#line 766

{
	
#line 768
do {
#line 768
  push(env, (STKVAL) (greylist_class[greylist_semantics].gl_pred(env, email)));
#line 768
  return;
#line 768
} while (0);
}
}

void
#line 772
bi_db_name(eval_environ_t env)
#line 772

#line 772 "bi_db.m4"
{
#line 772
	
#line 772

#line 772

#line 772
char * fmtid;
#line 772
        
#line 772
get_string_arg(env, 0, &fmtid);
#line 772
        
#line 772

#line 772
        adjust_stack(env, 1);
#line 772

#line 772
	if (prog_trace_option & BUILTIN_MASK_db)
#line 772
		prog_trace(env, "db_name %s",fmtid);;
#line 772

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 775
			env_throw_bi(env, mfe_not_found, "db_name", _("no such db format: %s"),fmtid)
#line 777
;
	do {pushs(env, fmt->dbname);
#line 778
return;
#line 778
} while (0);
}
}

void
#line 782
bi_db_get_active(eval_environ_t env)
#line 782

#line 782 "bi_db.m4"
{
#line 782
	
#line 782

#line 782

#line 782
char * fmtid;
#line 782
        
#line 782
get_string_arg(env, 0, &fmtid);
#line 782
        
#line 782

#line 782
        adjust_stack(env, 1);
#line 782

#line 782
	if (prog_trace_option & BUILTIN_MASK_db)
#line 782
		prog_trace(env, "db_get_active %s",fmtid);;
#line 782

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 785
			env_throw_bi(env, mfe_not_found, "db_get_active", _("no such db format: %s"),fmtid)
#line 787
;
	
#line 788
do {
#line 788
  push(env, (STKVAL) (fmt->enabled));
#line 788
  return;
#line 788
} while (0);
}
}

void
#line 792
bi_db_set_active(eval_environ_t env)
#line 792

#line 792 "bi_db.m4"
{
#line 792
	
#line 792

#line 792

#line 792
char * fmtid;
#line 792
        long  active;
#line 792
        
#line 792
get_string_arg(env, 0, &fmtid);
#line 792
        get_numeric_arg(env, 1, &active);
#line 792
        
#line 792

#line 792
        adjust_stack(env, 2);
#line 792

#line 792
	if (prog_trace_option & BUILTIN_MASK_db)
#line 792
		prog_trace(env, "db_set_active %s %lu",fmtid, active);;
#line 792

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 795
			env_throw_bi(env, mfe_not_found, "db_set_active", _("no such db format: %s"),fmtid)
#line 797
;
	fmt->enabled = active;
}
}

void
#line 802
bi_db_expire_interval(eval_environ_t env)
#line 802

#line 802 "bi_db.m4"
{
#line 802
	
#line 802

#line 802

#line 802
char * fmtid;
#line 802
        
#line 802
get_string_arg(env, 0, &fmtid);
#line 802
        
#line 802

#line 802
        adjust_stack(env, 1);
#line 802

#line 802
	if (prog_trace_option & BUILTIN_MASK_db)
#line 802
		prog_trace(env, "db_expire_interval %s",fmtid);;
#line 802

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 805
			env_throw_bi(env, mfe_not_found, "db_expire_interval", _("no such db format: %s"),fmtid)
#line 807
;
	
#line 808
do {
#line 808
  push(env, (STKVAL) (fmt->expire_interval));
#line 808
  return;
#line 808
} while (0);
}
}


#line 812
void
#line 812
db_init_builtin()
#line 812
{
#line 812
	#line 69 "bi_db.m4"
install_pragma("dbprop", 3, 4, _pragma_dbprop);
#line 198 "bi_db.m4"
va_builtin_install_ex("dbmap", bi_dbmap, 0, dtype_number, 3, 1, 0|0, dtype_string, dtype_string, dtype_number);
#line 208 "bi_db.m4"
va_builtin_install_ex("dbget", bi_dbget, 0, dtype_string, 4, 2, 0|0, dtype_string, dtype_string, dtype_string, dtype_number);
#line 219 "bi_db.m4"
va_builtin_install_ex("dbput", bi_dbput, 0, dtype_unspecified, 5, 2, 0|0, dtype_string, dtype_string, dtype_string, dtype_number, dtype_number);
#line 257 "bi_db.m4"
va_builtin_install_ex("dbdel", bi_dbdel, 0, dtype_unspecified, 4, 2, 0|0, dtype_string, dtype_string, dtype_number, dtype_number);
#line 323 "bi_db.m4"
DBTAB_id = builtin_priv_register(alloc_db_tab, destroy_db_tab,
#line 323
NULL);
#line 339 "bi_db.m4"
va_builtin_install_ex("dbfirst", bi_dbfirst, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 367 "bi_db.m4"
va_builtin_install_ex("dbnext", bi_dbnext, 0, dtype_number, 1, 0, 0|0, dtype_number);
#line 388 "bi_db.m4"
va_builtin_install_ex("dbkey", bi_dbkey, 0, dtype_string, 1, 0, 0|0, dtype_number);
#line 406 "bi_db.m4"
va_builtin_install_ex("dbvalue", bi_dbvalue, 0, dtype_string, 1, 0, 0|0, dtype_number);
#line 444 "bi_db.m4"
install_pragma("greylist", 2, 2, _pragma_greylist);
#line 490 "bi_db.m4"
	builtin_variable_install("greylist_seconds_left", dtype_number, SYM_VOLATILE, &greylist_seconds_left_loc);
#line 754 "bi_db.m4"
va_builtin_install_ex("greylist", bi_greylist, 0, dtype_number, 2, 0, 0|0, dtype_string, dtype_number);
#line 766 "bi_db.m4"
va_builtin_install_ex("is_greylisted", bi_is_greylisted, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 772 "bi_db.m4"
va_builtin_install_ex("db_name", bi_db_name, 0, dtype_string, 1, 0, 0|0, dtype_string);
#line 782 "bi_db.m4"
va_builtin_install_ex("db_get_active", bi_db_get_active, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 792 "bi_db.m4"
va_builtin_install_ex("db_set_active", bi_db_set_active, 0, dtype_unspecified, 2, 0, 0|0, dtype_string, dtype_number);
#line 802 "bi_db.m4"
va_builtin_install_ex("db_expire_interval", bi_db_expire_interval, 0, dtype_number, 1, 0, 0|0, dtype_string);

#line 812 "bi_db.m4"
	
#line 812
}
#line 812


#line 627 "../mfd/snarf.m4"

#line 627

